/**
 * @file dwarfutil.cpp
 * @author Martin Wagner
 *
 * @brief Implementation of some often used utility functions
 *
 * $Id: dwarfutil.cpp,v 1.1 2004/05/03 09:35:00 wagnerm Exp $
 * $Revision: 1.1 $
 */

// import TNT
#define TNT_BOUNDS_CHECK
#include "tnt.h"
using namespace TNT;


#include <assert.h>

#ifdef WIN32
# pragma warning(disable: 4786)
# include <windows.h>
#else
# include <unistd.h>
#endif

#include <math.h>

#ifdef WIN32
# pragma warning(disable: 4786)
# include <windows.h>
# include <time.h>
# include <sys/timeb.h>
#else
# include <unistd.h>
#endif

#include "math_library.h"


struct timeval currentTime() {

   struct timeval tvnow;

#ifdef _MSC_VER
   // Windows time
   struct _timeb wintime;
   _ftime( &wintime );
   tvnow.tv_sec = wintime.time;
   tvnow.tv_usec = wintime.millitm * 1000;
#else
   // Unix time
   gettimeofday( &tvnow, 0 );
#endif

   return tvnow;
}


/**
 * implementation based on proposal by Horn, idea:
 * first determine the largest entry of unit quaternion,
 * then use this to get other entries
 */
double* matrixToQuaternion( const double *m, double *q ) 
{
	// get entry of q with largest absolute value
	// note: we compute here 4 * q[..]^2 - 1
	double tmp[4];
	tmp[QW] = m[0] + m[5] + m[10];
	tmp[QX] = m[0] - m[5] - m[10];
	tmp[QY] = -m[0] + m[5] - m[10];
	tmp[QZ] = -m[0] - m[5] + m[10];
	int max = QW;
	if( tmp[QX] > tmp[max] ) max = QX;
	if( tmp[QY] > tmp[max] ) max = QY;
	if( tmp[QZ] > tmp[max] ) max = QZ;

	// depending on largest entry compute the other values
	// note: these formulae can be derived very simply from the
	//       matrix representation computed in quaternionToMatrix
	switch(max) {
	case QW:
		q[QW] = sqrt(tmp[QW]+1) * 0.5;
		q[QX] = (m[9] - m[6]) / ( 4 * q[QW] );
		q[QY] = (m[2] - m[8]) / ( 4 * q[QW] );
		q[QZ] = (m[4] - m[1]) / ( 4 * q[QW] );
		break;
	case QX:
		q[QX] = sqrt(tmp[QX]+1) * 0.5;
		q[QW] = (m[9] - m[6]) / ( 4 * q[QX] );
		q[QY] = (m[4] + m[1]) / ( 4 * q[QX] );
		q[QZ] = (m[2] + m[8]) / ( 4 * q[QX] );
		break;
	case QY:
		q[QY] = sqrt(tmp[QY]+1) * 0.5;
		q[QW] = (m[2] - m[8]) / ( 4 * q[QY] );
		q[QX] = (m[4] + m[1]) / ( 4 * q[QY] );
		q[QZ] = (m[9] + m[6]) / ( 4 * q[QY] );
		break;
	case QZ:
		q[QZ] = sqrt(tmp[QZ]+1) * 0.5;
		q[QW] = (m[4] - m[1]) / ( 4 * q[QZ] );
		q[QX] = (m[2] + m[8]) / ( 4 * q[QZ] );
		q[QY] = (m[9] + m[6]) / ( 4 * q[QZ] );
		break;
	default:
		assert(0);
		break;
	}
	
	checkUnit(q);

	return q;
}


double* axisAngleToQuaternion( double* axisAngle, double* quat ) 
{
	double length = sqrt( axisAngle[0]*axisAngle[0] + axisAngle[1]*axisAngle[1] 
												+ axisAngle[2]*axisAngle[2] );
	if( length < Q_EPSILON ) {
		quat[QW] = 0; quat[QX] = 0; quat[QY] = 1.0; quat[QZ] = 0;
		return quat;
	}
	double ahalf = axisAngle[3]/2.0;
	double s = sin( ahalf );
	quat[QW] = cos( ahalf );
	quat[QX] = s * axisAngle[0] / length;
	quat[QY] = s * axisAngle[1] / length;
	quat[QZ] = s * axisAngle[2] / length;

	checkUnit(quat);

	return quat;
}


double* quaternionToAxisAngle( double* q, double* axisAngle ) 
{
	checkUnit(q);

	axisAngle[0] = q[QX];
	axisAngle[1] = q[QY];
	axisAngle[2] = q[QZ];
	axisAngle[3] = 2 * acos( q[QW] );
	
	return axisAngle;
}



double* invertQuaternion( double *q ) 
{
	double norm = 1.0 / (q[QX]*q[QX] + q[QY]*q[QY] + q[QZ]*q[QZ] + q[QW]*q[QW]);
	q[QX] *= -norm;
	q[QY] *= -norm;
  q[QZ] *= -norm;
	return q;
}


double* conjugateQuaternion( double *q ) 
{
	q[QX] = -q[QX];
	q[QY] = -q[QY];
	q[QZ] = -q[QZ];
	return q;
}


double* normalizeQuaternion( double *q ) 
{
	double norm = sqrt( 1.0 / (q[QX]*q[QX] + q[QY]*q[QY] + q[QZ]*q[QZ] + q[QW]*q[QW]) );
	q[QX] *= norm;
	q[QY] *= norm;
	q[QZ] *= norm;
	q[QW] *= norm;
	return q;
}


double* multiplyQuaternion( double *q, const double *q1, const double *q2 ) 
{
	q[QX] = q1[QW]*q2[QX] + q1[QX]*q2[QW] + q1[QY]*q2[QZ] - q1[QZ]*q2[QY];
	q[QY] = q1[QW]*q2[QY] - q1[QX]*q2[QZ] + q1[QY]*q2[QW] + q1[QZ]*q2[QX];
	q[QZ] = q1[QW]*q2[QZ] + q1[QX]*q2[QY] - q1[QY]*q2[QX] + q1[QZ]*q2[QW];
	q[QW] = q1[QW]*q2[QW] - q1[QX]*q2[QX] - q1[QY]*q2[QY] - q1[QZ]*q2[QZ];
	return q;
}


/**
 * implementation based on precomputation, correct mult
 * matrix can be found in Horn, Closed-form solution of
 * absolute orientation using unit quaternions. (1987)
 */
double* rotateQuaternion( double *r, const double *q, const double *p )
{
	checkUnit(q);

	// precomputation of some values
	double xy = q[QX]*q[QY];
	double xz = q[QX]*q[QZ];
	double yz = q[QY]*q[QZ];
	double ww = q[QW]*q[QW];
	double wx = q[QW]*q[QX];
	double wy = q[QW]*q[QY];
	double wz = q[QW]*q[QZ];

	r[0] = p[0] * ( 2*(q[QX]*q[QX] + ww) - 1 ) + p[1] * 2 * (xy - wz) + p[2] * 2 * (wy + xz);
	r[1] = p[0] * 2 * (xy + wz) + p[1] * ( 2*(q[QY]*q[QY] + ww) - 1 ) + p[2] * 2 * (yz - wx);
	r[2] = p[0] * 2 * (xz - wy) + p[1] * 2 * (wx + yz) + p[2] * ( 2*(q[QZ]*q[QZ] + ww) - 1 );

	return r;
}


double* rotateMatrix( double* rotPoint, double *matrix, const double *point )
{
	Array2D<double> mat(4,4,matrix);
	Array2D<double> vec(4,1,1.0);
	vec[0][0] = point[0];
	vec[1][0] = point[1];
	vec[2][0] = point[2];

	Array2D<double> mult = matmult( mat, vec );

	rotPoint[0] = mult[0][0];
	rotPoint[1] = mult[1][0];
	rotPoint[2] = mult[2][0];
	return rotPoint;

}



void invertQuaternionPose( double* newQuat, double* newTrans, const double* oldQuat, const double* oldTrans )
{
	checkUnit(oldQuat);

	newQuat[0] = oldQuat[0];
	newQuat[1] = oldQuat[1];
	newQuat[2] = oldQuat[2];
	newQuat[3] = oldQuat[3];
	conjugateQuaternion(newQuat);

	rotateQuaternion( newTrans, newQuat, oldTrans );
	newTrans[0] = -newTrans[0];
	newTrans[1] = -newTrans[1];
	newTrans[2] = -newTrans[2];

	return;
}

double* negateQuaternion( double *quat )
{
	quat[QX] = -quat[QX];
	quat[QY] = -quat[QY];
	quat[QZ] = -quat[QZ];
	quat[QW] = -quat[QW];

	return quat;
}

double* quaternionNegateIfCloser( double* quat, const double* otherquat )
{
	double fDiff = 0.0;
	for(int i = 0; i < 4; i++ )
		fDiff += ( quat[i] - otherquat[i] ) * ( quat[i] - otherquat[i] );

	if ( fDiff > 2.0 )
		negateQuaternion( quat );
	return quat;
}




